#ifndef COVERAGEHYPOTHESIS_H_
#define COVERAGEHYPOTHESIS_H_

#include <deque>
#include <vector>
#include <algorithm>
#include <list>
#include <set>

#include "../etc/Typedef.h"
#include "../etc/StaticData.h"

#include "../Model/LatticeManager.h"

#include "../DataStructure/Coverage.h"

#include "./LexicalHypothesis.h"

using namespace std;

typedef vector<LexicalHypothesis*>       LexicalHypothesisStack;

class LexHypCmp{
 public:
  inline bool operator()(const LexicalHypothesis* a, const LexicalHypothesis* b) const {
    if(a->getLMState() < b->getLMState())                  return true;
    if(a->getLastCoveredSrcPos() < b->getLastCoveredSrcPos())  return true;
    /* // recombineɖړIꑤt[Ylꍇ
       if(a->getLastPhrasePair() < b->getLastPhrasePair())        return true;
    */
    return false;
  }
};

class LexHypCmpByScore{
 public:
  inline bool operator()(const LexicalHypothesis* a, const LexicalHypothesis* b) const {
    return a->getEstimatedScore() > b->getEstimatedScore();
  }
};

class CoverageHypothesis{
  typedef set<LexicalHypothesis*, LexHypCmp>          LHS;

 protected:
   Coverage coverage;
 private:
   static unsigned  beamWidth;
   static float     beamThreshold;
   static const LatticeManager* ptLatticeManager;

   float min;
   float max;
   
   unsigned size;

   LexicalHypothesisStack stk;
   LHS lexHypSet;
   

   void updateMaxScore(const float& score){
     max = score;
     min = std::min(min, max+beamThreshold);
   }

 public:
   CoverageHypothesis(const Coverage& c = Coverage())
     : min(-10000.0F), max(-10000.0F), coverage(c), size(0U), stk(StaticData::getInstance().getParameter().getBeamWidth()*2+10)
     {}

  const Coverage& getCoverage() const { return coverage; }

  const unsigned getStackSize() const { return size; }

  const float getEstimatedScore() const { return max; }
  const float getMaxScore() const { return max; }
  const float getMinScore() const { return min; }

  const float getMaxScoreSoFar() const { return stk[0]->getScore(); }

  const bool isTooBad(const float& score) const{
    return (score < min);
  }

  const LexicalHypothesis& operator[] (const unsigned nth) const{
    return *(stk[nth]);
  }

  const LexicalHypothesis& getNthHypothesis(const unsigned nth) const {
    return *(stk[nth]);
  }

  // pruningp
  bool operator>(const CoverageHypothesis& rhs) const{
    return max > rhs.getMaxScore();
  }

  // For CoverageHypothesis set
  bool operator<(const CoverageHypothesis& rhs) const{
    return coverage < rhs.coverage;
  }

  bool push(LexicalHypothesis* lexHypo){
    if(StaticData::getInstance().getParameter().getVerbose() >= 2)
      lexHypo->print();

    // recombine
    pair<LHS::iterator, bool> retVal = lexHypSet.insert(lexHypo);  
    LHS::iterator& equivHypoIter = retVal.first;
    bool bNewHypo = retVal.second;

    bool survived;
    if( bNewHypo ){
      stk[size++] = lexHypo;
      survived = true;
    }else{
      survived = false;

      LexicalHypothesis* equivHypo = *equivHypoIter;
      if(equivHypo->getEstimatedScore() < lexHypo->getEstimatedScore()){
	equivHypo->assign(lexHypo);
	// lexHypPool.push(lexHypo);
      }else{
	// lexHypPool.push(lexHypo);
	return survived;
      }
    }

    float score = lexHypo->getEstimatedScore();
    if(score > max) updateMaxScore(score);
  
    return survived;
  }

  void prune(const float minScore) {
    min = minScore;

    /*
    // for debug
    cerr << "coverage: "; coverage.print();
    cerr << "before coverage stack: (threshold = " << min << ")" << endl;
    for(unsigned i=0; i<size; i++)
      cerr << i << "th score : " << stk[i]->getEstimatedScore() << endl;
    cerr << "stack_end" << endl << endl;
    */
    /*
    cerr << "cov hyp size: " << size << endl;
    cerr << "threshold: " << min << endl;
    */

    LexicalHypothesis** end = &stk[size];
    size = 0;

    for(LexicalHypothesis** iter = &stk[0]; iter != end; iter++){
      // cerr << (*iter)->getEstimatedScore() << " (" << (*iter)->getID() << ")" << endl;
      if((*iter)->getEstimatedScore() < minScore){
        // lexHypPool.push(*iter);
      }else{
	stk[size++] = *iter;
      }
}
    
    /* // for debug
    cerr << "pruned coverage stack: (threshold = " << min << ")" << endl;
    coverage.print();
    for(unsigned i=0; i<size; i++){
      cerr << i << "th score : " << stk[i]->getEstimatedScore();
      cerr << endl;
    }
    cerr << "stack_end" << endl << endl;
    */

    LHS::iterator iter = lexHypSet.begin();
    while(iter != lexHypSet.end()){
      if((*iter)->getEstimatedScore() < minScore)
	lexHypSet.erase(iter++);
      else 
	iter++;
    }
    
    sort(&stk[0], &stk[size], LexHypCmpByScore());
  }

  virtual ~CoverageHypothesis(){}

  static void setBeamWidth(const unsigned _beamWidth) { beamWidth = _beamWidth; }
  static void setBeamThreshold(const float _beamThreshold){ beamThreshold = _beamThreshold; }
  static void setPtLatticeManeger(const LatticeManager* _ptLatticeManager){ ptLatticeManager = _ptLatticeManager; }
};

unsigned              CoverageHypothesis::beamWidth     = 1024;
float                 CoverageHypothesis::beamThreshold = log(0.00001F);
const LatticeManager* CoverageHypothesis::ptLatticeManager = NULL;

#endif /* COVERAGEHYPOTHESIS_H_ */
