#ifndef _TRANSLATOR_
#define _TRANSLATOR_

#include <iostream>
#include <deque>
#include <set>
#include <vector>
#include <algorithm>
#include <queue>
#include <list>

#include "./CoverageHypothesis.h"
#include "./LexicalHypothesis.h"

#include "../Model/LatticeManager.h"
#include "../Model/ReorderingModel.h"

#include "../DataStructure/ObjectPool.h"
#include "../etc/Typedef.h"

using namespace std;

typedef set<CoverageHypothesis>           CoverageHypothesisMap;
typedef vector<const CoverageHypothesis*> CoverageHypothesisStack;
typedef vector<CoverageHypothesisStack>   CoverageHypothesisStackArray;

class LexicalHypothesis;

class CovHypCmpByScoreSoFar{
 public:
  inline bool operator()(const CoverageHypothesis* a, const CoverageHypothesis* b) const {
    return a->getMaxScore() > b->getMaxScore();
  }
};

class Translator{
private:
  const Parameter&      param;
  LanguageModel&        lm;

  float maxScore;
  float minScore;
  float beamThreshold;
  unsigned beamWidth;

  const unsigned                 sentenceLength;
  const LatticeManager&          ltcManager;
  CoverageHypothesisStackArray   stackArray;
  CoverageHypothesisMap          covHypoMap[maxSentenceLength+1];

  unsigned lexHypNum;

  void initialize(){
    lexHypPool.clear();

    Coverage emptyCoverage;
    CoverageHypothesis& covHypo = const_cast<CoverageHypothesis&>(
        *(covHypoMap[0].insert(CoverageHypothesis(emptyCoverage)).first)
    );
    stackArray[0].push_back( &covHypo );
    LexicalHypothesis emptyHypo(&covHypo.getCoverage(), 0.0, 0.0);
    LexicalHypothesis* ptrEmptyHyp = lexHypPool.pop();
    *ptrEmptyHyp = emptyHypo;
    covHypo.push( ptrEmptyHyp );
    for(unsigned i=1; i<=sentenceLength; i++)
      stackArray[i].reserve(beamWidth);
  }

  const string& convertHypoArrayToString(const deque<const LexicalHypothesis*>& lexHypoArray) const{
    static string str = "";
    str.clear();

    for(unsigned i=0; i<lexHypoArray.size(); i++)
      str += lexHypoArray[i]->getLastPhrasePair().getTgtPhrase().toString();

    return str;
  }

  float processSentenceEnd(const unsigned end) const {
    return param.getWeightL() * lm.getLMScoreForSentenceEnd() - param.getWeightD()[0] * ReorderingModel::calcLinearDistPenalty(end, sentenceLength);
  }

  void updateMaxScore(const float score){
    maxScore = score;
    if(minScore < maxScore + beamThreshold)
      minScore = maxScore + beamThreshold;
  }

  bool isTooBad(const float& score, const CoverageHypothesis& covHypo) const {
    return score < minScore;
    // return (score < minScore || covHypo.isTooBad(score));
  }

  bool isTooBad(const float& score, const CoverageHypothesis* dummy) const {
    return score < minScore;
  }

  void pruneCardinality(LexicalHypothesis** lexHypStack, unsigned& lexHypNumOfCardinality, CoverageHypothesisMap& coverageHypoMap){
    // calc threshold score by stack size
    priority_queue<float> pq;
    for(LexicalHypothesis** iter = lexHypStack; iter != &lexHypStack[lexHypNumOfCardinality]; iter++){
      pq.push( (*iter)->getEstimatedScore() );
    }
    for(unsigned i=0; i<beamWidth; i++){
      pq.pop();
    }
    minScore = pq.top();
    
    // pruning for cardinality stack
    LexicalHypothesis** end = lexHypStack + lexHypNumOfCardinality;
    lexHypNumOfCardinality = 0; 
    for(LexicalHypothesis** iter = lexHypStack; iter < end; iter++){
      if( (*iter)->getEstimatedScore() < minScore ){
	// do nothing
      }else{
	lexHypStack[lexHypNumOfCardinality++] = *iter;
      }
    }

    // pruning for coverage stack
    CoverageHypothesisMap::iterator iter = coverageHypoMap.begin();
    while( iter != coverageHypoMap.end() ){
      if(iter->getStackSize())
	(const_cast<CoverageHypothesis&>(*iter)).prune(minScore);
      iter++;
    }
  }

  void pruneCardinality2(const unsigned cardinality, const LexicalHypothesis** lexHypStack, unsigned& lexHypNumOfCardinality, CoverageHypothesisMap& coverageHypoMap){
    CoverageHypothesisStack& currentStack = stackArray[cardinality];
    if(lexHypNumOfCardinality > beamWidth){
      priority_queue<float> pq;
      const LexicalHypothesis** const end = lexHypStack + lexHypNumOfCardinality;
      for(const LexicalHypothesis** iter = lexHypStack; iter != end; iter++)
	pq.push((*iter)->getEstimatedScore());
      for(unsigned i=0; i<beamWidth; i++) pq.pop();
      minScore = pq.top();
    }
    
    CoverageHypothesisMap::const_iterator iter = coverageHypoMap.begin();
    while( iter != coverageHypoMap.end() ){
      if(iter->getMaxScore() < minScore)
	coverageHypoMap.erase(iter++);
      else {
        const_cast<CoverageHypothesis&>(*iter).prune(minScore);
        currentStack.push_back(&(*iter));
	iter++;
      }
    }

    sort(currentStack.begin(), currentStack.end(), CovHypCmpByScoreSoFar());
  }

  void expandCardinality(const unsigned cardinality){
    maxScore = -10000.0F;
    minScore = -20000.0F;
    unsigned lexHypNumOfCardinality = 0;
    LexicalHypothesis* lexHypStack[10000] = { NULL };
  
    CoverageHypothesisMap& coverageHypoMap = covHypoMap[cardinality];
    
    const unsigned curMaxPhraseLength = (maxPhraseLength > cardinality) ? cardinality : maxPhraseLength;

    for(unsigned srcPhraseLength = 1; srcPhraseLength <= curMaxPhraseLength; srcPhraseLength++){
      unsigned prevCardinality = cardinality - srcPhraseLength;
      for(CoverageHypothesisStack::iterator covHypoIndex = stackArray[prevCardinality].begin();
          covHypoIndex != stackArray[prevCardinality].end();
          covHypoIndex++){
        const CoverageHypothesis& covHypo  = **covHypoIndex;
        const Coverage&           coverage = covHypo.getCoverage();
	
        const unsigned startPosMax = sentenceLength - srcPhraseLength;
        for(unsigned start = 0; start <= startPosMax; start++){
          unsigned end = start+srcPhraseLength-1;
          const Coverage& newCoveredCoverage = ltcManager.getCoverageLattice().get(end, start);

          if( coverage.isOverlap(newCoveredCoverage) ) continue;

          const Coverage& newHypoCoverage = coverage | newCoveredCoverage;
	  CoverageHypothesis* newCovHypo = NULL;

	  const unsigned size = covHypo.getStackSize();
          for(unsigned lexHypoIndex = 0; lexHypoIndex < size; lexHypoIndex++){
            const LexicalHypothesis& lexHypo = covHypo[lexHypoIndex];

            if(isTooBad(lexHypo.getScore(), newCovHypo))  break;

            const float distPenalty    = ReorderingModel::calcLinearDistPenalty(lexHypo.getLastCoveredSrcPos(), start);
            const float partialScore = lexHypo.getScore() - param.getWeightD()[DIST_PENALTY] * distPenalty;
	    
	    const float futureDistPenalty = ReorderingModel::calcFutureLinearDistPenalty(newHypoCoverage, cardinality, end);
            const float futureScore     = ltcManager.calcFutureScore(newHypoCoverage) + (param.isFutureDistScoreUse() ? (-param.getWeightD()[DIST_PENALTY] * futureDistPenalty) : 0.0F);
	    const float& bestTMScore    = (*ltcManager.getPhraseLattice().get(end, start).first).getScore();
              
            const PTableIndexPair& ptableIndexPair = ltcManager.getPhraseLattice().get(end, start);
            for(PhrasePairIndex iter = ptableIndexPair.first; iter!= ptableIndexPair.second; iter++){
              const PhrasePair& newAppliedPhrasePair = *iter;
              
              const float tmScore = newAppliedPhrasePair.getScore();
              
              const float c = partialScore + tmScore + futureScore;
              if( isTooBad(c, newCovHypo) )  break;
              
              float rmScore = 0;
              if(param.getRMFeatureNum()==MSDBIFE_FEATURE_NUM-1){
                rmScore += calcLexicalReorderingScore(newAppliedPhrasePair, start, end, lexHypo.getLastPhrasePair(), lexHypo.getLastCoveredSrcPos());
              }
	      
              if( isTooBad(c+rmScore, newCovHypo) )
                continue;
	          
              const bool unkWordFlag = (newAppliedPhrasePair.getTgtPhrase().getLMWordIDArray()[0] == -1);
              const LMWordIDArray& newAppliedLMWordIDArray = newAppliedPhrasePair.getTgtPhrase().getLMWordIDArray();
              const LMWordIDArray& prevLMHistory = lexHypo.getLMHistory();
              
              
              const float lmScore = lm.getLMScore(newAppliedLMWordIDArray, prevLMHistory);
              
              
              float Score = partialScore + tmScore + param.getWeightL() * lmScore + unkWordFlag * LanguageModel::unkWordPenalty ;
	          
              if(param.getRMFeatureNum()==MSDBIFE_FEATURE_NUM-1){
                const float rmScore = calcLexicalReorderingScore(newAppliedPhrasePair, start, end, lexHypo.getLastPhrasePair(), lexHypo.getLastCoveredSrcPos());
                Score += rmScore;
              }
              if(cardinality == sentenceLength) Score += processSentenceEnd(end);
              
              const float c2 = Score + futureScore;
              if( isTooBad(c2, newCovHypo) )  continue;
              
              if(maxScore < c2) updateMaxScore(c2);

	          lexHypNum++;
	          LexicalHypothesis* newLexHyp = lexHypPool.pop();
	      
	          if(!newCovHypo){
		         CoverageHypothesisMap::iterator newCovHypoIter = coverageHypoMap.insert( CoverageHypothesis(newHypoCoverage) ).first;
		         newCovHypo = const_cast<CoverageHypothesis*>(&(*newCovHypoIter));
              }
	      
	          newLexHyp->assign(
                  &newCovHypo->getCoverage(), Score, futureScore,
                  &newAppliedPhrasePair,
                  lm.getLastHistory(), lm.getLastHistoryEnd(),
		          lm.getLastState(),
                  end, &lexHypo
              );
	          if( newCovHypo->push(newLexHyp) ){
		         lexHypStack[lexHypNumOfCardinality++] = newLexHyp;
              }

              if(lexHypNumOfCardinality >= beamWidth*2){
                pruneCardinality(lexHypStack, lexHypNumOfCardinality, coverageHypoMap);
              }
            } // end of for(iter)
          } // end of for(lexHypoIndex)
        }// end of for(start)

      }// end of for(covHypoIndex)
    }// end of for(srcPhraseLength)

    pruneCardinality2(cardinality, (const LexicalHypothesis**)lexHypStack, lexHypNumOfCardinality, coverageHypoMap);

  }// end of expandCardinality()
  
  const float calcLexicalReorderingScore(const PhrasePair& newPhrasePair, const unsigned newStart, const unsigned newEnd,const PhrasePair& lastPhrasePair, const unsigned lastEnd){
    float rmScore = 0.0;
    if(lastEnd == -1){
      if(newStart == 0)
        rmScore += newPhrasePair.getRMScoreArray()[0];
      else
        rmScore += newPhrasePair.getRMScoreArray()[2];
    }else{
      if(lastEnd+1 == newStart){
        rmScore += newPhrasePair.getRMScoreArray()[0];
        rmScore += lastPhrasePair.getRMScoreArray()[3];
      }else if(newEnd == (lastEnd-lastPhrasePair.getSrcPhrase().size()+1)){
        rmScore += newPhrasePair.getRMScoreArray()[1];
        rmScore += lastPhrasePair.getRMScoreArray()[4];
      }else{
        rmScore += newPhrasePair.getRMScoreArray()[2];
        rmScore += lastPhrasePair.getRMScoreArray()[5];
      }
    }
    
    return rmScore;
  
  }

  void expandCoverage(){}

public:
  Translator(const LatticeManager& ltcManager)
   : sentenceLength(ltcManager.getPhraseLattice().size()),
     ltcManager(ltcManager),
     stackArray(sentenceLength+1),
     param(StaticData::getInstance().getParameter()),
     lm(StaticData::getInstance().getLanguageModel()),
     minScore(-20000.0F),
     maxScore(-10000.0F),
    beamThreshold(log(param.getBeamThreshold())),
    beamWidth(param.getBeamWidth())
     { initialize(); }

  
  const deque<const LexicalHypothesis*>& traceBack(const LexicalHypothesis* ptLexicalHypothesis) const {
    static deque<const LexicalHypothesis*> lexHypoPtrArray;
    lexHypoPtrArray.clear();

    for(const LexicalHypothesis* ptLexHypo = ptLexicalHypothesis;
         ptLexHypo->getPrevLexicalHypothesis() != NULL;
         ptLexHypo = ptLexHypo->getPrevLexicalHypothesis())
    {
      lexHypoPtrArray.push_front(ptLexHypo);
      if(StaticData::getInstance().getParameter().getVerbose() > 0) ptLexHypo->print();
    }// end of for

    return lexHypoPtrArray;
  }

  const deque<const LexicalHypothesis*>& traceBack() const {
    static deque<const LexicalHypothesis*> lexHypoPtrArray;
    lexHypoPtrArray.clear();

    for(const LexicalHypothesis* ptLexHypo = &stackArray.back().front()->getNthHypothesis(0);
         ptLexHypo->getPrevLexicalHypothesis() != NULL;
         ptLexHypo = ptLexHypo->getPrevLexicalHypothesis())
    {
      lexHypoPtrArray.push_front(ptLexHypo);
      if(StaticData::getInstance().getParameter().getVerbose() > 0) ptLexHypo->print();
    }// end of for

    return lexHypoPtrArray;
  }


  const string& translate(){
    lexHypNum = 0;

    for(int cardinality = 1; cardinality <= sentenceLength; cardinality++){
      expandCardinality(cardinality);
      
      // for debug
      if(param.getVerbose() > 1){
	const CoverageHypothesisStack& currentStack = stackArray[cardinality];
	cerr << "survived hypthesis of cardinality " << cardinality << ": " << endl;
	for(unsigned i=0; i<currentStack.size(); i++)
          for(unsigned j=0; j<currentStack[i]->getStackSize(); j++)
	    cerr << (*currentStack[i])[j].getID() << "(" << (*currentStack[i])[j].getEstimatedScore() << ")" << endl;
	cerr << "cardinality" << cardinality << "end" << endl;
      }
    }

    const LexicalHypothesis& bestHypothesis = stackArray.back().front()->getNthHypothesis(0);

    return convertHypoArrayToString( traceBack(&bestHypothesis) );
  }

  const float getBestScore() const {
    return stackArray.back().front()->getNthHypothesis(0).getEstimatedScore();
  }

  const string& getNthTranslation(const unsigned n) const {
    const LexicalHypothesis& nthHypothesis = stackArray.back().front()->getNthHypothesis(n);
    return convertHypoArrayToString( traceBack(&nthHypothesis) );
  }
  
  unsigned getExpandedLexHypNum() const { return lexHypNum; }

};

#endif
