#ifndef _PHRASE_TABLE_
#define _PHRASE_TABLE_

#include <vector>
#include <deque>
#include <iostream>
#include <map>
#include <algorithm>
#include <string>
#include <utility>
#include <sstream>
#include <numeric>
#include <queue>


#include "../DataStructure/Phrase.h"
#include "../DataStructure/PhrasePair.h"
#include "../etc/Utility.h"
#include "ReorderingModel.h"

using namespace std;

typedef vector<PhrasePair>                     PhrasePairArray;
typedef PhrasePairArray::const_iterator        PhrasePairIndex;
// typedef const PhrasePair*                    PhrasePairIndex;
typedef pair<PhrasePairIndex, PhrasePairIndex>   PTableIndexPair;
typedef map<Phrase, PTableIndexPair>                 PhraseMap;
typedef pair<Phrase, PTableIndexPair>                PhraseMapEntry;
typedef deque<unsigned>                        SizeArray;


class PhraseTable{
 private:
  PhrasePairArray  phrasePairArray;
  PhraseMap        phraseMap;
  unsigned         ttableLimit;
  float            ttableThreshold;
  string           delimiter;
  SizeArray        tgtVariationNumArray;
  const Parameter* param;

 private:
  void       printProgress(const unsigned currentNum, const unsigned tableSize) const {
    static unsigned prev;
    unsigned progress = (unsigned)( (float)currentNum/tableSize *100 );
    if(progress != 0 && progress > prev){
      cerr << ".";
      if( progress % 10 == 0 && progress % 100 != 0 )
	cerr << progress/10 << "0% ->" << endl;
      else if( progress % 100 == 0 )
	cerr << progress/10 << "0% //" << endl;
    }
    prev = progress;
  }

  PhrasePair strToPhrasePair(const string& line, const Parameter& param, LanguageModel& lm) const {
    StringArray elems = split(line, param.getDelimiter());

    string src(elems[0]);
    string tgt(elems[1]);

    // skip word alignment(elems[2,3]) in phrase

    istringstream iStrStream(elems[4]);
    const unsigned tmFeatureNum = param.getTMFeatureNum();
    ScoreArray tmScoreArray(tmFeatureNum);
    for(unsigned i=0; i<tmFeatureNum; i++)
      iStrStream >> tmScoreArray[i];
    tmScoreArray = log(tmScoreArray);
    elems.pop_front();


    ScoreArray rmScoreArray(-0.1, 1);// param.getRMFeatureNum());

    const unsigned tgtPhraseSize = split(tgt).size();
    const float wordPenalty = param.getWeightW() * (-(float)tgtPhraseSize);

    const float score        = inner_product(&tmScoreArray[0], &tmScoreArray[0]+tmFeatureNum, &param.getWeightT()[0], 0.0F) + wordPenalty;
    const float lmScore      = lm.getLMScore( Phrase(tgt).getLMWordIDArray() );
    const float futureScore  = score + param.getWeightL()*lmScore;

    return PhrasePair(src, tgt, score, futureScore, tmScoreArray, rmScoreArray);
  }
  
   PhrasePair strToPhrasePair(const string& line, const string& rline, const Parameter& param, LanguageModel& lm) const {

    StringArray elems = split(line, param.getDelimiter());

    string src(elems[0]);
    string tgt(elems[1]);

    // skip word alignment(elems[2,3]) in phrase

    istringstream iStrStream(elems[4]);
    const unsigned tmFeatureNum = param.getTMFeatureNum();
    ScoreArray tmScoreArray(tmFeatureNum);
    for(unsigned i=0; i<tmFeatureNum; i++)
      iStrStream >> tmScoreArray[i];
    tmScoreArray = log(tmScoreArray);
    elems.pop_front();
    
    
    
    StringArray rElems = split(rline, param.getDelimiter());
    
    string rSrc(rElems[0]);
    string rTgt(rElems[1]);
    
    if(src != rSrc || tgt != rTgt){
      cerr << "PT line: " << line << endl;
      cerr << "RT line: " << rline << endl;
      cerr << "Not match Phrase Pair!!" << endl;
      exit(1);
    }
    
    
    istringstream iRStrStream(rElems[2]);
    const unsigned rmFeatureNum = param.getRMFeatureNum();
    ScoreArray rmScoreArray(rmFeatureNum);
    for(unsigned j=0; j<rmFeatureNum; j++){
      iRStrStream >> rmScoreArray[j];
    }
    rmScoreArray = log(rmScoreArray);
    for(int j = 0; j<rmFeatureNum; j++){
      rmScoreArray[j] = rmScoreArray[j]*param.getWeightD()[j+1];
    }
    rElems.pop_front();

    const unsigned tgtPhraseSize = split(tgt).size();
    const float wordPenalty = param.getWeightW() * (-(float)tgtPhraseSize);

    const float score        = inner_product(&tmScoreArray[0], &tmScoreArray[0]+tmFeatureNum, &param.getWeightT()[0], 0.0F) + wordPenalty;
    const float lmScore      = lm.getLMScore( Phrase(tgt).getLMWordIDArray() );
    const float futureScore  = score + param.getWeightL()*lmScore;

    return PhrasePair(src, tgt, score, futureScore, tmScoreArray, rmScoreArray);
  }

  unsigned checkTTableSize(const string& ttableFileName) {
    ifstream ttableFile(ttableFileName.c_str());
    unsigned ttableSize = 0;
    unsigned allPhraseNum = 0;

    unsigned currentSrcPhraseNum = 0;
    string   currentSrcPhrase;
    string   line;

    cerr << "Check phrase table size" << endl;
    while( !ttableFile.eof() ){
      getline(ttableFile, line);
      if(line.empty()){
        if( !ttableFile.eof() )  cerr << "Caution: empty line exists in ttableFile, #line:" << allPhraseNum << endl;
        continue;
      }

      allPhraseNum++;
      if( allPhraseNum   % (100*1000)  == 0 )
        cerr << ".";
      if( allPhraseNum   % (1000*1000) == 0 )
        cerr << allPhraseNum/(1000*1000) << "M" << endl;

      string srcPhrase = split(line, " "+delimiter+" ")[0];
      if(srcPhrase == currentSrcPhrase){
        if(currentSrcPhraseNum < ttableLimit)  currentSrcPhraseNum++;
      }else{
        ttableSize += currentSrcPhraseNum;
        currentSrcPhrase = srcPhrase;
        currentSrcPhraseNum = 1;
      }
    }
    cerr << endl;

    cerr << "# all  phrase: " << allPhraseNum << endl;
    cerr << "# load phrase: " << ttableSize;
    cerr << "(" << (float)ttableSize/allPhraseNum*100 << "%)" << endl << endl;

    ttableFile.close();
    return ttableSize;
  }

  void load(const Parameter& param, const string& ttableFileName, const unsigned ttableSize, LanguageModel& lm){
    this->param = &param;

    phrasePairArray.reserve((unsigned)(ttableSize*1.2));
    cerr << "reserved phrase table size: " << phrasePairArray.capacity() << endl;
    cerr << "load phrase table" << endl;
    ifstream ttableFile(ttableFileName.c_str());
    typedef multiset<PhrasePair, greater<PhrasePair> > EquivarentPhraseSet;

    EquivarentPhraseSet equivPhraseSet;
    Phrase currentSrcPhrase;
    string line;
    unsigned num = 0;

    while( !ttableFile.eof() ){
      getline(ttableFile, line);    if( line.empty() ) continue;

      printProgress(num, ttableSize);

      PhrasePair phrasePair = strToPhrasePair(line, param, lm);

      // for debug
      if(StaticData::getInstance().getParameter().getVerbose() > 3)
        cerr << phrasePair << endl;

      const Phrase& src = phrasePair.getSrcPhrase();
      if(src.size() > maxPhraseLength) maxPhraseLength = src.size();

      if(src == currentSrcPhrase){
        equivPhraseSet.insert(phrasePair);
      }else{
	EquivarentPhraseSet::const_iterator start = equivPhraseSet.begin();
	EquivarentPhraseSet::const_iterator end   = equivPhraseSet.end();

	float threshold = -10000.0F;
	if(equivPhraseSet.size() > ttableLimit){
	  // check threshold
	  priority_queue<float> bestFutureScores;
	  for(EquivarentPhraseSet::const_iterator iter = start; iter != end; iter++)
	    bestFutureScores.push(iter->getFutureScore());
	  for(unsigned i=0; i<ttableLimit; i++)
	    bestFutureScores.pop();
	  threshold = bestFutureScores.top();
	}
	// push pp
	PhrasePairArray::const_iterator s = phrasePairArray.end();
	unsigned size = 0;
	for(EquivarentPhraseSet::const_iterator iter = start; size < ttableLimit && iter != end; iter++)
	  if(iter->getFutureScore() > threshold){
	    phrasePairArray.push_back(*iter);
	    size++;
	  }
	PhrasePairArray::const_iterator e = phrasePairArray.end();

        const PhraseMap::const_iterator it = phraseMap.insert( PhraseMapEntry(Phrase(currentSrcPhrase), PTableIndexPair(s, e)) ).first;

        num += distance(s, e);

        equivPhraseSet.clear();
        equivPhraseSet.insert(phrasePair);
        currentSrcPhrase = src;
      }

    }// end of while

    printProgress(num, ttableSize);
    cerr << "# loaded phrase: " << num << endl;
    cerr << "complete loading phrase table!!"  << endl;
    cerr << endl;
  }
  
  void load(const Parameter& param, const string& ttableFileName, const unsigned ttableSize, const string& rtableFileName, LanguageModel& lm){
    this->param = &param;

    phrasePairArray.reserve((unsigned)(ttableSize*1.2));
    cerr << "reserved phrase table size: " << phrasePairArray.capacity() << endl;
    cerr << "load phrase table" << endl;
    ifstream ttableFile(ttableFileName.c_str());
    ifstream rtableFile(rtableFileName.c_str());
    typedef multiset<PhrasePair, greater<PhrasePair> > EquivarentPhraseSet;

    EquivarentPhraseSet equivPhraseSet;
    Phrase currentSrcPhrase;
    string line;
    string rline;
    unsigned num = 0;

    while( !ttableFile.eof() && !rtableFile.eof()){
      getline(ttableFile, line);    if( line.empty() ) continue;
      getline(rtableFile, rline);   if( rline.empty()) continue;
      printProgress(num, ttableSize);

      PhrasePair phrasePair = strToPhrasePair(line, rline, param, lm);

      // for debug
      if(StaticData::getInstance().getParameter().getVerbose() > 3)
        cerr << phrasePair << endl;

      const Phrase& src = phrasePair.getSrcPhrase();
      if(src.size() > maxPhraseLength) maxPhraseLength = src.size();

      if(src == currentSrcPhrase){
        equivPhraseSet.insert(phrasePair);
      }else{
	EquivarentPhraseSet::const_iterator start = equivPhraseSet.begin();
	EquivarentPhraseSet::const_iterator end   = equivPhraseSet.end();

	float threshold = -10000.0F;
	if(equivPhraseSet.size() > ttableLimit){
	  // check threshold
	  priority_queue<float> bestFutureScores;
	  for(EquivarentPhraseSet::const_iterator iter = start; iter != end; iter++)
	    bestFutureScores.push(iter->getFutureScore());
	  for(unsigned i=0; i<ttableLimit; i++)
	    bestFutureScores.pop();
	  threshold = bestFutureScores.top();
	}
	// push pp
	PhrasePairArray::const_iterator s = phrasePairArray.end();
	unsigned size = 0;
	for(EquivarentPhraseSet::const_iterator iter = start; size < ttableLimit && iter != end; iter++)
	  if(iter->getFutureScore() > threshold){
	    phrasePairArray.push_back(*iter);
	    size++;
	  }
	PhrasePairArray::const_iterator e = phrasePairArray.end();

        const PhraseMap::const_iterator it = phraseMap.insert( PhraseMapEntry(Phrase(currentSrcPhrase), PTableIndexPair(s, e)) ).first;

        num += distance(s, e);

        equivPhraseSet.clear();
        equivPhraseSet.insert(phrasePair);
        currentSrcPhrase = src;
      }

    }// end of while

    printProgress(num, ttableSize);
    cerr << "# loaded phrase: " << num << endl;
    cerr << "complete loading phrase table!!"  << endl;
    cerr << endl;
  }

 public:
  explicit PhraseTable(const Parameter& param, LanguageModel& lm)
    :ttableLimit(param.getTTableLimit()), ttableThreshold( param.getTTableThreshold() ), delimiter(param.getDelimiter())
  {
    string ttableFileName(param.getTTableFileName());
    unsigned ttableSize = checkTTableSize(ttableFileName);
    if(param.getRMFeatureNum()==0){
      load(param, ttableFileName,ttableSize, lm);
    }else{
      string rtableFileName(param.getRTableFileName());
      load(param, ttableFileName, ttableSize, rtableFileName, lm);
    }
  }

  bool isUnknownPhrase(const Phrase& srcPhrase) const {
    return phraseMap.end() == phraseMap.find(srcPhrase);
  }

  const PTableIndexPair& set(const Phrase& srcPhrase) {
    const string phrase = strip(srcPhrase.toString());
    const ScoreArray tmScoreArray(0.0, StaticData::getInstance().getParameter().getTMFeatureNum());

    ScoreArray rmScoreArray(-0.1, StaticData::getInstance().getParameter().getRMFeatureNum());
    if(StaticData::getInstance().getParameter().getRMFeatureNum()==0){
      rmScoreArray.resize(1,0.0);
    }
    const PhrasePair newPhrasePair(phrase, phrase, (float)0.0, (1.0F+param->getWeightL())*LanguageModel::unkWordPenalty, tmScoreArray, rmScoreArray);

    const PhrasePairIndex start= phrasePairArray.end();
    phrasePairArray.push_back(newPhrasePair);
    const PhrasePairIndex end = phrasePairArray.end();

    const PTableIndexPair indexPair = PTableIndexPair(start, end);
    return ( *( phraseMap.insert( PhraseMapEntry(srcPhrase, indexPair) ).first ) ).second;
  }

  const PTableIndexPair& get(const Phrase& srcPhrase) const {
    return (*phraseMap.find(srcPhrase)).second;
  }


  const PTableIndexPair getUnkIndexPair() const {
    return PTableIndexPair(phrasePairArray.end(), phrasePairArray.end());
  }

  void print() const {
    for(unsigned i=0; i<phrasePairArray.size(); i++)
      cerr << phrasePairArray[i].toString() << endl;
  }
};

#endif
